use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::time::Duration;
use tokio::runtime::Runtime;
use actix_web::{
    dev::ServerHandle,
    rt,
    web,
    App as WebApp,
    web::Data,
    web::Query,
    HttpRequest,
    HttpResponse,
    HttpServer,
    Responder};
use std::sync::{Arc, Mutex};
use std::collections::HashMap;
use std::any::Any;
use serde_derive::Deserialize;
use reqwest::blocking;
use actix_web_actors::ws;
use actix::prelude::*;
use std::{sync::OnceLock};


use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;

pub trait UserApp {
    fn get_pkg_name(&self) -> String;
    fn init(&self);
    fn update(&self);
    fn uninit(&self);
    fn get_http_listen_port(&self) -> i32;
}

#[derive(SharedFromSelf)]
pub struct App {
    _self_wp: Weak::<RefCell::<Self>>,
    zone_id: i32,
    node_id: i32,
    instance_id: i32,
    //tokio_rt: Runtime,
    user_app: Option<&'static dyn UserApp>,
    im_msgs: Rc::<RefCell::<r9::ListHead::<IMMsgNode>>>,
    im_work_msgs: Rc::<RefCell::<r9::ListHead::<IMMsgNode>>>,
    im_mutex: Mutex<i32>,
}

#[derive(Default)]
struct IMMsgNode {
    msg_id: u16,
    args: Option<Vec<Rc::<dyn Any>>>,
    cb: Option<Box::<dyn FnMut (&Option<Vec<Rc::<dyn Any>>>)>>,
    entry: Rc::<RefCell::<r9::ListHead::<IMMsgNode>>>,
}

impl App {

    pub fn instance() -> Rc::<RefCell::<Self>> {
        static mut _INSTANCE: Option<Rc::<RefCell::<App>>> = None;
        unsafe {
            match &_INSTANCE {
                Some(v) => {
                    return v.clone();
                }
                None => {
                    _INSTANCE = Some(Rc::new(RefCell::new(
                        App{
                            zone_id: 0,
                            node_id: 0,
                            instance_id: 0,
                            //tokio_rt: Runtime::new().unwrap(),
                            user_app: None,
                            im_msgs: r9::ListHead::<IMMsgNode>::new_head(),
                            im_work_msgs: r9::ListHead::<IMMsgNode>::new_head(),
                            im_mutex: Mutex::new(1),
                            _self_wp: Default::default(),
                        }
                    )));
                    _INSTANCE.clone().unwrap().borrow_mut()._self_wp =
                        Rc::downgrade(&_INSTANCE.clone().unwrap());
                }
            }
            return _INSTANCE.clone().unwrap().clone();
        }
    }

    pub fn init(user_app: &'static dyn UserApp) {
        App::instance().borrow_mut().user_app = Some(user_app);
        //App::instance().borrow_mut().tokio_rt.enter();
        //App::instance().borrow_mut().init_http_server();
        crate::Timer::instance().borrow_mut().init();
        user_app.init();
    }

    pub fn uninit() {
        let user_app = App::instance().borrow_mut().user_app.clone();
        //user_app.as_ref().unwrap().borrow_mut().uninit();
        crate::Timer::instance().borrow_mut().uninit();
    }

    pub fn run() {
        loop {
            //crate::Timer::update();
            App::dispatch_immsg();
            App::update_user_app();
            std::thread::sleep(Duration::from_millis(1));
        }
    }

    pub fn get_pkg_name(&self) -> String {
        return self.user_app.unwrap().get_pkg_name();
    }

    pub fn new_uuid(&self) -> i64 {
        return 0;
    }

    pub fn get_instance_id(&self) -> i32 {
        return self.instance_id;
    }

    pub fn get_zone_id(&self) -> i32 {
        return self.zone_id;
    }

    pub fn get_node_id(&self) -> i32 {
        return self.node_id;
    }

    pub fn has_flag(&self) -> bool {
        return false;
    }

    pub fn get_now_second(&self) -> i64 {
        return 0;
    }

    pub fn get_now_millis(&self) -> i64 {
        return 0;
    }

    pub fn get_time_offset(&self) -> i64 {
        return 0;
    }

    pub fn add_immsg(&mut self, msg_id: u16, args: Option<Vec<Rc::<dyn Any>>>,
                     cb: Box::<dyn FnMut (&Option<Vec<Rc::<dyn Any>>>)>) {
        let node = IMMsgNode{
            msg_id: msg_id,
            args: args,
            cb: Some(cb),
            entry: Default::default(),
        };
        self.im_mutex.lock();
        r9::ListHead::add_tail(&self.im_msgs, &node.entry);
    }

    fn dispatch_immsg() {
        let im_work_msgs = App::instance().borrow_mut().fetch_immsg();
        while !im_work_msgs.borrow().empty() {
            {
                let node = &im_work_msgs.borrow().first_entry();
                match &mut node.upgrade().unwrap().borrow_mut().cb {
                    Some(v) => {
                        (*v)(&node.upgrade().unwrap().borrow_mut().args);
                    }
                    None => {
                    }
                }
                node.upgrade().unwrap().borrow_mut().entry.borrow_mut().del_init();
            }
        }
    }

    fn update_user_app() {
        App::instance().borrow_mut().user_app.unwrap().update();
    }

    fn fetch_immsg(&mut self) -> Rc::<RefCell::<r9::ListHead::<IMMsgNode>>> {
        r9::ListHead::replace_init(&self.im_msgs, &self.im_work_msgs);
        return self.im_work_msgs.clone();
    }

}
