use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::sync::Mutex;
//use futures::lock::Mutex;
use std::sync::Arc;
use std::thread;
use std::thread::Thread;
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use std::default::Default;

pub struct Queue<T> {
    msg_list: Mutex<Rc::<RefCell::<crate::ListHead<T>>>>,
    pub work_list: Mutex<Rc::<RefCell::<crate::ListHead<T>>>>,
}

impl<T> Queue<T> {

    pub fn new() -> Rc::<RefCell::<Self>> {
        let this = Rc::new(RefCell::new(Self{
            msg_list: Mutex::new(crate::ListHead::<T>::new_head()),
            work_list: Mutex::new(crate::ListHead::<T>::new_head()),
        }));
        return this;
    }

    pub fn new_ex() -> Arc::<Mutex::<Self>> {
        let this = Arc::new(Mutex::new(Self{
            msg_list: Mutex::new(crate::ListHead::<T>::new_head()),
            work_list: Mutex::new(crate::ListHead::<T>::new_head()),
        }));
        return this;
    }

    pub fn push(&self, node: &Rc::<RefCell::<crate::ListHead<T>>>) {
        crate::ListHead::<T>::add_tail(&self.msg_list.lock().unwrap(), node);
    }

    pub fn fetch(&self) {
        if !self.msg_list.lock().unwrap().borrow().empty() &&
            self.work_list.lock().unwrap().borrow().empty() {
                crate::ListHead::replace_init(&self.msg_list.lock().unwrap(), &self.work_list.lock().unwrap());
        }
    }

    pub fn empty(&self) -> bool {
        if !self.work_list.lock().unwrap().borrow().empty() {
            return false
        }
        return self.msg_list.lock().unwrap().borrow().empty()
    }

}
