use std::sync::Mutex;
use std::sync::Arc;
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use std::default::Default;

pub struct QueueLock<T> {
    msg_list: Mutex<Arc::<Mutex::<crate::ListHeadLock<T>>>>,
    pub work_list: Mutex<Arc::<Mutex::<crate::ListHeadLock<T>>>>,
}

unsafe impl<T> Send for QueueLock<T> {}
unsafe impl<T> Sync for QueueLock<T> {}

impl<T> QueueLock<T> {

    pub fn new() -> Arc::<Mutex::<Self>> {
        let this = Arc::new(Mutex::new(Self{
            msg_list: Mutex::new(crate::ListHeadLock::<T>::new_head()),
            work_list: Mutex::new(crate::ListHeadLock::<T>::new_head()),
        }));
        return this;
    }

    pub fn new_ex() -> Arc::<Mutex::<Self>> {
        let this = Arc::new(Mutex::new(Self{
            msg_list: Mutex::new(crate::ListHeadLock::<T>::new_head()),
            work_list: Mutex::new(crate::ListHeadLock::<T>::new_head()),
        }));
        return this;
    }

    pub fn push(&self, node: &Arc::<Mutex::<crate::ListHeadLock<T>>>) {
        crate::ListHeadLock::<T>::add_tail(&self.msg_list.lock().unwrap(), node);
    }

    pub fn fetch(&self) -> &Self {
        if !self.msg_list.lock().unwrap().lock().unwrap().empty() &&
            self.work_list.lock().unwrap().lock().unwrap().empty() {
                crate::ListHeadLock::replace_init(&self.msg_list.lock().unwrap(), &self.work_list.lock().unwrap());
        }
        return self
    }

    pub fn empty(&self) -> bool {
        if !self.work_list.lock().unwrap().lock().unwrap().empty() {
            return false
        }
        return self.msg_list.lock().unwrap().lock().unwrap().empty()
    }

}
