use proc_macro::TokenStream;
use quote::quote;
use syn;

#[cfg(test)]
mod tests {
    #[test]
    fn it_works() {
        let result = 2 + 2;
        assert_eq!(result, 4);
    }
}

#[proc_macro_derive(SharedFromSelf, attributes(helper))]
pub fn shared_from_self_derive(input: TokenStream) -> TokenStream {
    let ast = syn::parse(input).unwrap();

    // Build the trait implementation
    impl_shared_from_self_macro(&ast)
}

fn impl_shared_from_self_macro(ast: &syn::DeriveInput) -> TokenStream {
    let name = &ast.ident;
    let (impl_generics, _, _) = ast.generics.split_for_impl();
    let gen = quote! {
        impl #impl_generics SharedFromSelf for #name #impl_generics {
            fn shared_from_self(&self) -> Rc::<RefCell::<Self>> {
                return self._self_wp.upgrade().unwrap();
            }
        }
    };
    gen.into()
}

#[proc_macro_derive(SharedFromSelfLock, attributes(helper))]
pub fn shared_from_self_lock_derive(input: TokenStream) -> TokenStream {
    let ast = syn::parse(input).unwrap();

    // Build the trait implementation
    crate::impl_shared_from_self_lock_macro(&ast)
}

fn impl_shared_from_self_lock_macro(ast: &syn::DeriveInput) -> TokenStream {
    let name = &ast.ident;
    let (impl_generics, _, _) = ast.generics.split_for_impl();
    let gen = quote! {
        impl #impl_generics SharedFromSelfLock for #name #impl_generics {
            fn shared_from_self_lock(&self) -> Arc::<Mutex::<Self>> {
                return self._self_wp.upgrade().unwrap();
            }
        }
    };
    gen.into()
}

#[proc_macro_derive(Singleton)]
pub fn singleton_derive(input: TokenStream) -> TokenStream {
    let ast = syn::parse(input).unwrap();

    // Build the trait implementation
    impl_singleton_macro(&ast)
}

fn impl_singleton_macro(ast: &syn::DeriveInput) -> TokenStream {
    let name = &ast.ident;
    let (_, _, _) = ast.generics.split_for_impl();
    let gen = quote! {
        impl #name {
            pub fn instance() -> Rc::<RefCell::<Self>> {
                static mut _INSTANCE: Option<Rc::<RefCell::<#name>>> = None;
                unsafe {
                    match &_INSTANCE {
                        Some(v) => {
                            return v.clone();
                        }
                        None => {
                            _INSTANCE = Some(Rc::new(RefCell::new(Self::new())));
                            _INSTANCE.clone().unwrap().borrow_mut()._self_wp =
                                Rc::downgrade(&_INSTANCE.clone().unwrap());
                        }
                    }
                    return _INSTANCE.clone().unwrap().clone();
                }
            }
        }
    };
    //println!("{}", gen);
    gen.into()
}
